// deployer main programm
// Copyright (C) 2024 Jean-Cloud
// GNU General Public License v3

#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <cstring>
#include <filesystem>
#include <fstream>
#include "Services.h"
#include "Modules.h"
#include "BashManager.h"
using namespace std;

void help(char * argv0)
{
    //temporary 
    cout << "usage: ./" << argv0 <<"action file \n with action=deploy or remove and file=path to a file or all. \n" << endl;
}

int isServiceOnServer(string serviceUsername)
//this method tests if a certain service is on the current server
//it looks into the /etc/hosts file thanks to a pipe to a separate bash process
{   
    string cmd ="getent hosts " +serviceUsername;  
    string result = BashManager::executeAndReadResult(cmd);
    if(result.find("::1")!=string::npos){ //if result contains "::1" which is the notation for loopback in IpV6
        cout << "service on server" << endl;
        return 0;
    }
    cout << "service not on server" << endl;
    return 1;
   return 0;
}

int createUser(string serviceUsername)
{
    //this method creates a Unix user dedicated to the service
    //get the User ID from servers.csv
    int uidStart=2000; //so that the uids do not overlap with existing uids
    Services services = Services("./services/services.csv");
    const Service * service = services.findByUsername(serviceUsername);
    int uid = (*service).getUserID()+uidStart;
    //test if user already exists
    string cmd = "id -u "+serviceUsername; 
    string res = BashManager::executeAndReadResult (cmd);
    if(res==to_string(uid)){
        cout << "user already existed" << endl;
        return 0;
    }
    //create user
    string cmd2 ="useradd -u " + to_string(uid) + " " + serviceUsername + "&& usermod -s /sbin/nologin "+ serviceUsername; //no direct login
    string res2 = BashManager::executeAndReadResult(cmd2);
    if (res2 != ""){
        cerr << "Error when executing the bash command to create a user specific to the service." << endl;
        cerr << res2 << endl;
        return -1;
    }
    return 0;
}

int createEnvService(string serviceUsername)
{
    //create directories
    //filesystem::create_directories() 


    /*
    run mkdir -p "$DATA_DIR" "$HTTP_DIR"
	run chown $uid "$DATA_DIR"
	run chmod 751 "$DATA_DIR"
	run chown $uid:www-data -R "$HTTP_DIR"
	if [ -d "$SECRET_DIR" ] ; then
		run chown $uid "$SECRET_DIR" -R
		run chmod 751 "$SECRET_DIR" -R
	fi*/    return 0;
}

int removeEnvService()
{
    /*
        [ -d "$HTTP_DIR" ] && rm -r "$HTTP_DIR"*/
    return 0;
}

int createEnv()
{
    string proxyDir="etc/nginx";
    string dns_certs_path="/data/dnscerts.jean-cloud.org/certs/live";
    string http_certs_path="/etc/letsencrypt/live";

    ofstream outfile ("/etc/jeancloud.env");
    outfile << "proxy_dir=" << proxyDir << endl;
    outfile << "nginx_conf_path=" << proxyDir << "/sites-enabled/" <<endl;
    outfile << "new_nginx_conf_path=" << proxyDir << "/new-sites-enabled" << endl; 
    outfile << "dns_certs_path=" << dns_certs_path << endl;
    outfile << "http_certs_path=" << http_certs_path << endl;
    outfile << "dummy_cert_path=" << http_certs_path << "/dummy" <<endl;
    outfile << "servicefile=/services/services.csv";
    outfile << "services_uid_start=2000" << endl;
    outfile.close();
    return 0;
}

int deployAll()
{
    //this method deploys all the services that are on this server
    cout << "deploying all" <<endl;
    createEnv();
    //for each service deploy service
    return 0;
}

int deployService(string serviceUsername){
    //this method deploys indicated service if it is on this server
    if (isServiceOnServer(serviceUsername)==0){
        //bash user creation
        if(int userCreated = createUser(serviceUsername);userCreated!=0){
            return -1;
        }
        //environment variables creation
        if(int envCreated = createEnvService(serviceUsername);envCreated!=0){
            return -1; 
        }
        //call to the deploy functionality of all modules
        //the modules themselves determine their course of action depending on the service
        for(Module * mod_ptr : modules){
            int modResult = (*mod_ptr).deploy(serviceUsername);
            if (modResult!=0){
                cerr << "Error in " << (*mod_ptr) << " when deploying " << serviceUsername << endl;
            }
        }
        return 0;
        cerr << "Error creating user and environment of "<<serviceUsername << endl;

    }
    return 0; 
}

int removeAll()
{
    cout << "removing all"<<endl;
    return 0;
}

int removeService(string serviceUsername)
{
    cout<< "removing service"<<endl;
    return 0;
}

int main(int argc, char *argv[])
{
    //check number of arguments
    if (argc != 3){
        cerr << "Invalid number of arguments. \n" << endl;
        help(argv[0]);
        return -1;
    } else {
        //check that data is mounted on the server
        if (!(filesystem::exists("/data/mounted"))) {
            cerr << "Error. The data is not mounted on the server" << endl;
            return -1;
        } else {
            //read instructions
            string action = argv[1];
            string service = argv[2];
            if (action=="deploy"){
                if(service=="all"){
                    deployAll();
                } else {
                    deployService(service);
                }
            } else if (action=="remove"){
                if(service=="all"){
                    removeAll();
                } else {
                    removeService(service);
                }
            } else {
            cerr << "Invalid argument. \n" << endl;
            return -1;
            }
        }
    }
    return 0;
}


